@@ -107,18 +107,19 @@ int minibuffer_auto_raise;
 
 static Lisp_Object last_exact_completion;
 
-Lisp_Object Quser_variable_p;
-
-Lisp_Object Qminibuffer_default;
-
 /* Non-nil means it is the window for C-M-v to scroll
    when the minibuffer is selected.  */
 extern Lisp_Object Vminibuf_scroll_window;
 
 extern Lisp_Object Voverriding_local_map;
 
+Lisp_Object Quser_variable_p;
+
+Lisp_Object Qminibuffer_default;
+
 Lisp_Object Qcurrent_input_method, Qactivate_input_method;
 
+extern Lisp_Object Qmouse_face;
 
 /* Put minibuf on currently selected frame's minibuffer.
    We do this whenever the user starts a new minibuffer
@@ -369,6 +370,10 @@ read_minibuf (map, initial, prompt, backup_n, expflag,
     unbind_to (count1, Qnil);
   }
 
+  /* If appropriate, copy enable-multibyte-characters into the minibuffer.  */
+  if (inherit_input_method)
+    current_buffer->enable_multibyte_characters = enable_multibyte;
+
   /* Put in the initial input.  */
   if (!NILP (initial))
     {
@@ -387,10 +392,6 @@ read_minibuf (map, initial, prompt, backup_n, expflag,
   if (STRINGP (input_method) && !NILP (Ffboundp (Qactivate_input_method)))
     call1 (Qactivate_input_method, input_method);
 
-  /* If appropriate, copy enable-multibyte-characters into the minibuffer.  */
-  if (inherit_input_method)
-    current_buffer->enable_multibyte_characters = enable_multibyte;
-
   if (!NILP (current_buffer->enable_multibyte_characters)
       && ! STRING_MULTIBYTE (minibuf_prompt))
     minibuf_prompt = Fstring_make_multibyte (minibuf_prompt);
@@ -1820,7 +1821,9 @@ DEFUN ("display-completion-list", Fdisplay_completion_list, Sdisplay_completion_
 Each element may be just a symbol or string\n\
 or may be a list of two strings to be printed as if concatenated.\n\
 `standard-output' must be a buffer.\n\
-At the end, run the normal hook `completion-setup-hook'.\n\
+The actual completion alternatives, as inserted, are given `mouse-face'\n\
+properties of `highlight'.\n\
+At the end, this runs the normal hook `completion-setup-hook'.\n\
 It can find the completion buffer in `standard-output'.")
   (completions)
      Lisp_Object completions;
@@ -1928,11 +1931,44 @@ It can find the completion buffer in `standard-output'.")
 	  else if (!NILP (current_buffer->enable_multibyte_characters)
 		   && !STRING_MULTIBYTE (string))
 	    string = Fstring_make_multibyte (string);
-	  Fprinc (string, Qnil);
+
+	  if (BUFFERP (Vstandard_output))
+	    {
+	      XSETINT (startpos, BUF_PT (XBUFFER (Vstandard_output)));
+
+	      Fprinc (string, Qnil);
+
+	      XSETINT (endpos, BUF_PT (XBUFFER (Vstandard_output)));
+
+	      Fput_text_property (startpos, endpos,
+				  Qmouse_face, intern ("highlight"),
+				  Vstandard_output);
+	    }
+	  else
+	    {
+	      Fprinc (string, Qnil);
+	    }
 
 	  /* Output the annotation for this element.  */
 	  if (CONSP (elt))
-	    Fprinc (Fcar (Fcdr (elt)), Qnil);
+	    {
+	      if (BUFFERP (Vstandard_output))
+		{
+		  XSETINT (startpos, BUF_PT (XBUFFER (Vstandard_output)));
+
+		  Fprinc (Fcar (Fcdr (elt)), Qnil);
+
+		  XSETINT (endpos, BUF_PT (XBUFFER (Vstandard_output)));
+
+		  Fset_text_properties (startpos, endpos, Qnil,
+					Vstandard_output);
+		}
+	      else
+		{
+		  Fprinc (Fcar (Fcdr (elt)), Qnil);
+		}
+	    }
+
 
 	  /* Update COLUMN for what we have output.  */
 	  column += length;
